/*---------------------------------------------------------------------------------
|     Project | L3DTio_HFF 
| Description | A plugin for L3DT to implement the Heightfield file (HFF) format.
|      Author | Aaron Torpy (aaron@bundysoft.com)
|   Copyright | (C) Copyright 2006-2011 Aaron Torpy. All rights reserved.
|         URL | http://www.bundysoft.com/wiki/doku.php?id=plugins:fileio:L3DTio_HFF
\---------------------------------------------------------------------------------*/

#include "stdafx.h"
#include "L3DTio_HFF.h"


#include <Zeolite.h>

#include <zList.h>
#include <zVar.h>
#include <zmap.h>
#include <zFormat.h>
#include <zProgBox.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

// CL3DTio_HFFApp

BEGIN_MESSAGE_MAP(CL3DTio_HFFApp, CWinApp)
END_MESSAGE_MAP()


// CL3DTio_HFFApp construction

CL3DTio_HFFApp::CL3DTio_HFFApp() {
	// TODO: add construction code here,
	// Place all significant initialization in InitInstance
}


// The one and only CL3DTio_HFFApp object

CL3DTio_HFFApp theApp;


// CL3DTio_HFFApp initialization

BOOL CL3DTio_HFFApp::InitInstance() {
	CWinApp::InitInstance();

	return TRUE;
}

zeoexport bool __stdcall zeoInitPlugin() {

	CzFormat format;

	// heightfield
	if(format.Create("Heightfield", "L3DT Heightfield File", "hff", "L3DT")) {

		// set the flags
		format.SetFlags(ZFORMAT_FLAG_NATIVE|ZFORMAT_FLAG_MOSAIC|ZFORMAT_FLAG_LOAD|ZFORMAT_FLAG_SAVE|ZFORMAT_FLAG_LOADTILE|ZFORMAT_FLAG_SAVETILE); // native I/O and mosaic

		// set the options
		int DataSize = 2;
		bool FloatFlag = false;
		int BlockSize = 0;
		format.SetOptionValue("DataSize", VarID_int, &DataSize);
		format.SetOptionValue("TileSize", VarID_int, &BlockSize);
		format.SetOptionValue("FloatFlag", VarID_bool, &FloatFlag);
	}
	if(format.CreateGeneric(MAP_Heightfield, "L3DT Heightfield File", "hff", "L3DT")) {

		// set the flags
		format.SetFlags(ZFORMAT_FLAG_NATIVE|ZFORMAT_FLAG_MOSAIC|ZFORMAT_FLAG_LOAD|ZFORMAT_FLAG_SAVE|ZFORMAT_FLAG_LOADTILE|ZFORMAT_FLAG_SAVETILE); // native I/O and mosaic

		// set the options
		int DataSize = 2;
		bool FloatFlag = false;
		int BlockSize = 0;
		format.SetOptionValue("DataSize", VarID_int, &DataSize);
		format.SetOptionValue("TileSize", VarID_int, &BlockSize);
		format.SetOptionValue("FloatFlag", VarID_bool, &FloatFlag);
	}

	// water map
	if(format.Create("Water map", "L3DT Heightfield File", "hff", "L3DT")) {


		// set the flags
		format.SetFlags(ZFORMAT_FLAG_LOAD|ZFORMAT_FLAG_SAVE); // neither native nor mosaic

		// set the options
		int DataSize = 2;
		bool FloatFlag = false;
		int BlockSize = 0;
		format.SetOptionValue("DataSize", VarID_int, &DataSize);
		format.SetOptionValue("TileSize", VarID_int, &BlockSize);
		format.SetOptionValue("FloatFlag", VarID_bool, &FloatFlag);
	}

	// salinity map
	if(format.Create("Salinity map", "L3DT Heightfield File", "hff", "L3DT")) {


		// set the flags
		format.SetFlags(ZFORMAT_FLAG_NATIVE|ZFORMAT_FLAG_MOSAIC|ZFORMAT_FLAG_LOAD|ZFORMAT_FLAG_SAVE|ZFORMAT_FLAG_LOADTILE|ZFORMAT_FLAG_SAVETILE); // native I/O and mosaic

		// set the options
		int DataSize = 1;
		bool FloatFlag = false;
		int BlockSize = 0;
		format.SetOptionValue("DataSize", VarID_int, &DataSize);
		format.SetOptionValue("TileSize", VarID_int, &BlockSize);
		format.SetOptionValue("FloatFlag", VarID_bool, &FloatFlag);
	}

	//
	// add any further initialisation here
	//

	return true;
}


zeoexport bool __stdcall zeoSaveMapFile(ZMAP hMap, LPCSTR lpFileName, ZFORMAT hFormat, ZVAR hProgWnd) {

	CzMap zmap;
	zmap.Attach(hMap);

	CzFormat format;
	format.Attach(hFormat);

	CzProgBox prog;
	prog.Attach(hProgWnd);
	
	struct _iobuf *fh;
	int i, j, tx, ty;
	char TempBuf[10];
	USHORT TempUSI;
	int TempInt;
	CString TempStr;
	float TempFloat;
	BYTE TempByte;
	float HFpxl;
	float HFmin, HFmax;
	WaterMapPixel WMpxl;

	// check the map type
	long MapTypeID = zmap.GetMapType();
	switch(MapTypeID) {
	default:
		zeoReportError("L3DTio_HFF::SaveMapFile error:\r\n - unsupported map type");
		return false;
	case MAP_Heightfield:
	case MAP_WaterMap:
		// OK
		break;
	}
	
	// get the altitude range
	if(!zmap.GetMinMaxAlt(HFmin, HFmax)) {
		zeoReportError("L3DTio_HFF::SaveMapFile error:\r\n - call to GetMinMaxAlt failed");
		return false;
	}


	 // the file version number for HFF v1.0
	int FileVersionNo = DEF_FILE_HF300;

	// defaults
	int DataSize = 2;
	bool FloatFlag = false;
	int BlockSize = 0;
	format.GetOptionValue("DataSize", VarID_int, &DataSize);
	format.GetOptionValue("TileSize", VarID_int, &BlockSize);
	format.GetOptionValue("FloatFlag", VarID_bool, &FloatFlag);

	long nx = zmap.nx();
	long ny = zmap.ny();

	if(nx<=0 || ny<=0) {
		zeoReportError("L3DTio_HFF::SaveMapFile error:\r\n - map not initialised");
		return false;
	}
	
	// making file header
	FileHeader_HF300 FH_HF300;
	FH_HF300.nx = nx;
	FH_HF300.ny = ny;
	FH_HF300.TileSize = BlockSize;
	FH_HF300.WrapFlag = zmap.GetWrapFlag();
	FH_HF300.DataOffset = 64;			// 64 for HFF_v1.0
	FH_HF300.DataSize = DataSize;
	FH_HF300.FloatFlag = FloatFlag;
	FH_HF300.HorizScale = zmap.GetHorizScale();

	if(FH_HF300.FloatFlag) {
		switch(FH_HF300.DataSize) {
		case 4:
			FH_HF300.VertScale = 1.0f;
			FH_HF300.VertOffset = 0.0f;
			break;
		default:
			zeoReportError("L3DTio_HFF::SaveMapFile error:\r\n - unacceptable data size in HFFv1.0 (float)");
			return false;
		}
	} else {
		switch(FH_HF300.DataSize) {
		case 1:
			FH_HF300.VertScale = (float)(HFmax - HFmin)/255.0f;
			FH_HF300.VertOffset = HFmin;
			break;
		case 2: // USHORT
			FH_HF300.VertScale=(float)(HFmax - HFmin)/65535.0f;
			FH_HF300.VertOffset = HFmin;
			break;
		default:
			zeoReportError("L3DTio_HFF::SaveMapFile error:\r\n - unacceptable data size in HFFv1.0 (int)");
			return false;
		}
	}

	if(FH_HF300.TileSize>1) {
		if(FH_HF300.nx%FH_HF300.TileSize!=0 || FH_HF300.ny%FH_HF300.TileSize!=0) {
			zeoReportError("L3DTio_HFF::SaveMapFile error:\r\n - tile size is incompatible with map size in HFFv1.0");
			return false;
		}
	}
	
	// opening file for writing
	fh=fopen(lpFileName, "wb"); 
	if(fh==NULL) {
		CString ErrStr;
		ErrStr.Format("L3DTio_HFF::SaveMapFile error:\r\n - cannot open file '%s' for writing", lpFileName);
		zeoReportError(ErrStr);
		return false;
	}

	// writing header
	sprintf(TempBuf, "L3DT");
	fwrite(TempBuf, 4, 1, fh);
	fwrite((void*)&FileVersionNo, 2, 1, fh);
	sprintf(TempBuf, "HFF_v1.0");
	fwrite(TempBuf, 8, 1, fh);
	fwrite(&FH_HF300.DataOffset, 2, 1, fh);
	fwrite(&FH_HF300.nx, 4, 1, fh);
	fwrite(&FH_HF300.ny, 4, 1, fh);
	fwrite(&FH_HF300.DataSize, 1, 1, fh);
	fwrite(&FH_HF300.FloatFlag, 1, 1, fh);
	fwrite(&FH_HF300.VertScale, 4, 1, fh);
	fwrite(&FH_HF300.VertOffset, 4, 1, fh);
	fwrite(&FH_HF300.HorizScale, 4, 1, fh);
	fwrite(&FH_HF300.TileSize, 2, 1, fh);
	fwrite(&FH_HF300.WrapFlag, 1, 1, fh);
	TempBuf[0]=0;
	fwrite(TempBuf, 1, 23, fh);	// filling in reserved junk
	
	// preparing progress bar
	__int64 p, q, pmax, qstep;
	MSG TempMSG;
	if(hProgWnd) {
		prog.SetTitle("Saving HFF");
		prog.ShowWnd();
		pmax = (__int64)nx * (__int64)ny;
		qstep = (__int64)pmax/50; // HACK!
		if(qstep<1)	qstep=1;
		else 		pmax=50;
		q=0, p=0;
	}


	int TempTileSize = FH_HF300.TileSize;
	if(TempTileSize<1)
		TempTileSize = 1;
	
	bool ErrorFlag = false;

	int nty = ny / TempTileSize;
	int ntx = nx / TempTileSize;
	if(ny%TempTileSize)	nty++;
	if(nx%TempTileSize)	ntx++;

	int i2, j2;

	

	for(ty=0; ty<nty && !ErrorFlag; ty++) {
	for(tx=0; tx<ntx && !ErrorFlag; tx++) {
		for(j=0; j<TempTileSize && !ErrorFlag; j++) {
		for(i=0; i<TempTileSize && !ErrorFlag; i++) {

			i2 = i + tx*TempTileSize;
			j2 = j + ty*TempTileSize;

			if(i2<nx && j2<ny) {

				if(hProgWnd) {
					if(q==qstep)	{ p++; q=0; }
					if(q++==0) {
						PeekMessage(&TempMSG, NULL, 0, 0, PM_REMOVE);
						zprogbox_SetProgress(hProgWnd, p, pmax);
					}
				}

				switch(MapTypeID) {
				case MAP_Heightfield:
					if(!zmap.GetPixel(i2, j2, &HFpxl)) {
						zeoReportError("L3DTio_HFF::SaveMapFile error:\r\n - GetPixel returned false from HF");
						ErrorFlag = true;
					}
					break;
				case MAP_WaterMap:
					if(!zmap.GetPixel(i2, j2, &WMpxl)) {
						zeoReportError("L3DTio_HFF::SaveMapFile error:\r\n - GetPixel returned false from WM");
						ErrorFlag = true;
					}
					HFpxl = WMpxl.WaterLevel;
					break;
				}

				if(!ErrorFlag) {
					if(FH_HF300.FloatFlag) {
						TempFloat=(HFpxl - FH_HF300.VertOffset)/FH_HF300.VertScale; // note gain & offset are taken into account when saving floats
						fwrite((void*)&TempFloat, 4, 1, fh);
					} else {
						switch(FH_HF300.DataSize) {
						case 1:
							TempInt=(int)((HFpxl-FH_HF300.VertOffset)/FH_HF300.VertScale);
							if(TempInt<0)		TempInt=0;
							if(TempInt>255)		TempInt=255;
							TempByte=(BYTE)TempInt;
							fwrite((void*)&TempByte, 1, 1, fh);
							break;
						case 2:

							TempInt=(int)((HFpxl-FH_HF300.VertOffset)/FH_HF300.VertScale);
							if(TempInt<0)		TempInt=0;
							if(TempInt>65535)	TempInt=65535;
							TempUSI=(USHORT)TempInt;

							fwrite((void*)&TempUSI, 2, 1, fh);
							break;
						default:
							zeoReportError("L3DTio_HFF::SaveMapFile error\r\n - pixel size is invalid");
							ErrorFlag = true;
							break;
						}
					}
				}
			}
		}
		}
	}
	}

	if(hProgWnd) {
		prog.HideWnd();
	}

	fclose(fh);

	if(ErrorFlag) {
		zeoReportError("L3DTio_HFF::SaveMapFile error:\r\n - Save terminated following errors");
		return false;
	}
	
	return true;
}


zeoexport bool __stdcall zeoLoadMapFile(ZMAP hMap, LPCSTR lpFileName, long MapTypeID, ZFORMAT hFormat, ZVAR hProgWnd) {

	CzMap zmap;
	zmap.Attach(hMap);

	CzFormat format;
	format.Attach(hFormat);

	CzProgBox prog;
	prog.Attach(hProgWnd);
	
	struct _iobuf *fh;
	int i, j, tx, ty, i2, j2;
	char TempBuf[8];
	USHORT TempUSI;
	float TempFloat;
	CString TempStr;
	BYTE TempByte;
	char FileVersionCharBuf[9];
	char* CharBuf;


	USHORT FileVersionNo;

	// opening file for reading
	fh=fopen(lpFileName, "rb"); 
	if(fh==NULL) {
		TempStr.Format("File '%s' not found", lpFileName);
		zeoReportError(TempStr);
		return false;
	}

	// check the file header
	fread(TempBuf, 4, 1, fh);
	TempBuf[4] = 0; 	// null-terminating string
	if(strcmp(TempBuf, "L3DT")!=0) {
		// case for old maps that didn't have "L3DT" written at start
		zeoReportError("L3DTio_HFF::LoadMapFile error\r\n - invalid file type");
		fclose(fh);
		return false;
	}
		
	// read the file version number
	fread(&FileVersionNo, 2, 1, fh);
	switch(FileVersionNo) {
	default:
		zeoReportError("L3DTio_HFF::LoadMapFile error\r\n - unknown file version");
		fclose(fh);
		return false;
	case DEF_FILE_HF300:
		//this is OK
		break;
	}

	// reading ASCII external file version number
	fread(FileVersionCharBuf, 8, 1, fh);
	FileVersionCharBuf[8]=0;	// null-terminating string
	if(strcmp(FileVersionCharBuf, "HFF_v1.0")!=0) {
		TempStr.Format("L3DTio_HFF::LoadMapFile error\r\n - unknown file version '%s' in HF300", FileVersionCharBuf);
		zeoReportError(TempStr);
		fclose(fh);
		return false;
	}

	// declare a file header struct
	FileHeader_HF300 FH_HF300;

	// read the data offset
	fread(&FH_HF300.DataOffset, 2, 1, fh);

	// check data offset
	if(FH_HF300.DataOffset<64) {
		zeoReportError("L3DTio_HFF::LoadMapFile error:\r\n - Invalid file header length");
		fclose(fh);
		return false;
	}

	// reading remainder of header into char buffer
	CharBuf=(char*)calloc(1, FH_HF300.DataOffset-16);
	fread(CharBuf, FH_HF300.DataOffset-16, 1, fh);

	// copy from buffer into header struct
	memcpy(&FH_HF300.nx, CharBuf, 4);
	memcpy(&FH_HF300.ny, CharBuf+4, 4);
	memcpy(&FH_HF300.DataSize, CharBuf+8, 1);
	memcpy(&FH_HF300.FloatFlag, CharBuf+9, 1);
	memcpy(&FH_HF300.VertScale, CharBuf+10, 4);
	memcpy(&FH_HF300.VertOffset, CharBuf+14, 4);
	memcpy(&FH_HF300.HorizScale, CharBuf+18, 4);
	memcpy(&FH_HF300.TileSize, CharBuf+22, 2);
	memcpy(&FH_HF300.WrapFlag, CharBuf+24, 1);

	// free buffer
	free(CharBuf);

	// check tiling (is it compatible with map size?)
	if(FH_HF300.TileSize>1) {
		if(FH_HF300.nx%FH_HF300.TileSize!=0 || FH_HF300.ny%FH_HF300.TileSize!=0) {
			zeoReportError("L3DTio_HFF::LoadMapFile error\r\n - tile size is incompatible with map size in HF300");
			fclose(fh);
			return false;
		}
	}

	// check vertical scaling
	if(FH_HF300.VertScale<=0.0) {
		zeoReportError("L3DTio_HFF::LoadMapFile error\r\n - vertical scale is equal to or smaller than zero");
		fclose(fh);
		return false;
	}

	// check horizontal scaling
	if(FH_HF300.HorizScale<=0.0) {
		FH_HF300.HorizScale = 1.0;
//		theAPI..ReportError("L3DTio_HFF::LoadMapFile error\r\n - horizontal scale is equal to or smaller than zero");
//		zeoReportError("L3DTio_HFF::LoadMapFile error\r\n - horizontal scale is equal to or smaller than zero");
//		fclose(fh);
//		return false;
	}

	// checking data types
	if(FH_HF300.FloatFlag) {
		switch(FH_HF300.DataSize) {
		case 4:	// that's OK
			break;
		default:
			zeoReportError("L3DTio_HFF::LoadMapFile error\r\n - unacceptable data size (float)");
			fclose(fh);
			return false;
		}
	} else {
		switch(FH_HF300.DataSize) {
		case 1: // BYTE is OK
			break;
		case 2: // USHORT is OK
			break;
		default:
			zeoReportError("L3DTio_HFF::LoadMapFile error\r\n - unacceptable data size (int)");
			fclose(fh);
			return false;
		}
	}

	// freeing map memory if already allocated
	if(zmap.GetMapType()!=MAP_Empty)
		zmap.Free();

	long nx = FH_HF300.nx;
	long ny = FH_HF300.ny;

	if(nx<=0||ny<=0) {
		zeoReportError("L3DTio_HFF::LoadMapFile error\r\n - invalid map size");
		fclose(fh);
		return false;
	}

	// initialising map
	if(!zmap.Init(nx, ny, MAP_Heightfield, FH_HF300.HorizScale, (bool)FH_HF300.WrapFlag)) {
		zeoReportError("L3DTio_HFF::LoadMapFile error\r\n - cannot initialise map");
		fclose(fh);
		return false;
	}

	// preparing progress bar
	__int64 p, q, pmax, qstep;
	MSG TempMSG;
	if(hProgWnd) {
		prog.SetTitle("Loading HFF");
		prog.ShowWnd();
		pmax = nx * ny;
		qstep = (__int64)pmax/50; // HACK!
		if(qstep<1)	qstep=1;
		else 		pmax=50;
		q=0, p=0;
	}

	// reading data (handling both tiled and non-tiled maps)
	int TileSize = FH_HF300.TileSize;
	if(TileSize<=0)	TileSize = 1;

	// stepping through tiles
	bool ErrorFlag = false;
	for(ty=0; ty<ny/TileSize && !ErrorFlag; ty++) {
	for(tx=0; tx<nx/TileSize && !ErrorFlag; tx++) {

		// stepping through pixels in each tile
		for(j=0; j<TileSize && !ErrorFlag; j++) {
		for(i=0; i<TileSize && !ErrorFlag; i++) {

			// converting from local tile coords to global map coords
			i2 = i+tx*TileSize;
			j2 = j+ty*TileSize;

			if(i2<nx && j2<ny) {

				if(hProgWnd) {
					if(q==qstep)	{ p++; q=0; }
					if(q++==0) {
						PeekMessage(&TempMSG, NULL, 0, 0, PM_REMOVE);
						zprogbox_SetProgress(hProgWnd, p, pmax);
					}
				}

				if(FH_HF300.FloatFlag) {
					fread((void*)&TempFloat, 4, 1, fh);
					TempFloat = TempFloat*FH_HF300.VertScale+FH_HF300.VertOffset;
				} else {
					switch(FH_HF300.DataSize) {
					case 1:
						fread((void*)&TempByte, 1, 1, fh);
						TempFloat = (float)TempByte*FH_HF300.VertScale+FH_HF300.VertOffset;
						break;	
					case 2:
						fread((void*)&TempUSI, 2, 1, fh);
						TempFloat = (float)TempUSI*FH_HF300.VertScale+FH_HF300.VertOffset;
						break;
					default:
						ErrorFlag = true;
						zeoReportError("L3DTio_HFF::LoadMapFile error\n - pixel size is invalid in HF300");
						break;
					}
				}

				if(!zmap.SetPixel(i2, j2, &TempFloat)) {
					ErrorFlag = true;
					zeoReportError("L3DTio_HFF::LoadMapFile error\n - call to SetPixel failed");
				}
			}
		}
		}
	}
	}

	fclose(fh);

	if(hProgWnd) {
		prog.HideWnd();
	}

	if(ErrorFlag) {
		return false;
	}

	// updating file config	
	int datasize = FH_HF300.DataSize;
	int tilesize = FH_HF300.TileSize;
	bool floatflag = FH_HF300.FloatFlag;
	format.SetOptionValue("DataSize", VarID_int, &datasize);
	format.SetOptionValue("TileSize", VarID_int, &tilesize);
	format.SetOptionValue("FloatFlag", VarID_bool, &floatflag);

	return true;
}


zeoexport bool __stdcall zeoSaveTileFile(ZMAP hMap, LPVOID hTile, LPCSTR lpFileName, ZFORMAT hFormat) {

	CzMap zmap;
	zmap.Attach(hMap);

	CzFormat format;
	format.Attach(hFormat);
	
	struct _iobuf *fh;
	int i, j, tx, ty;
	char TempBuf[10];
	USHORT TempUSI;
	int TempInt;
	CString TempStr;
	float TempFloat;
	BYTE TempByte;
	float HFpxl;
	float HFmin, HFmax;

	// check the map type
	long MapTypeID = zmap.GetMapType();
	switch(MapTypeID) {
	default:
		zeoReportError("L3DTio_HFF::SaveTileFile error:\r\n - unsupported map type");
		return false;
	case MAP_Heightfield:
		// OK
		break;
	}
	
	// get the altitude range
	if(!zmap_tile_GetMinMaxAlt(hTile, HFmin, HFmax)) {
		zeoReportError("L3DTio_HFF::SaveTileFile error:\r\n - call to tile_GetMinMaxAlt failed");
		return false;
	}


	 // the file version number for HFF v1.0
	int FileVersionNo = DEF_FILE_HF300;

	// defaults
	int DataSize = 2;
	bool FloatFlag = false;
	int BlockSize = 0;
	format.GetOptionValue("DataSize", VarID_int, &DataSize);
	format.GetOptionValue("TileSize", VarID_int, &BlockSize);
	format.GetOptionValue("FloatFlag", VarID_bool, &FloatFlag);

	long nx, ny;
	nx = ny = zmap.GetTileSize();

	if(nx<=0 || ny<=0) {
		zeoReportError("L3DTio_HFF::SaveTileFile error:\r\n - invalid tile size");
		return false;
	}
	
	// making file header
	FileHeader_HF300 FH_HF300;
	FH_HF300.nx = nx;
	FH_HF300.ny = ny;
	FH_HF300.TileSize = BlockSize;
	FH_HF300.WrapFlag = zmap.GetWrapFlag();
	FH_HF300.DataOffset = 64;			// 64 for HFF_v1.0
	FH_HF300.DataSize = DataSize;
	FH_HF300.FloatFlag = FloatFlag;
	FH_HF300.HorizScale = zmap.GetHorizScale();

	if(FH_HF300.FloatFlag) {
		switch(FH_HF300.DataSize) {
		case 4:
			FH_HF300.VertScale = 1.0f;
			FH_HF300.VertOffset = 0.0f;
			break;
		default:
			zeoReportError("L3DTio_HFF::SaveTileFile error:\r\n - unacceptable data size in HFFv1.0 (float)");
			return false;
		}
	} else {
		switch(FH_HF300.DataSize) {
		case 1:
			FH_HF300.VertScale = (float)(HFmax - HFmin)/255.0f;
			FH_HF300.VertOffset = HFmin;
			break;
		case 2: // USHORT
			FH_HF300.VertScale=(float)(HFmax - HFmin)/65535.0f;
			FH_HF300.VertOffset = HFmin;
			break;
		default:
			zeoReportError("L3DTio_HFF::SaveTileFile error:\r\n - unacceptable data size in HFFv1.0 (int)");
			return false;
		}
	}

	if(FH_HF300.VertScale<=0.0f) {
		FH_HF300.VertScale = 1; 

		// supress error; this occurs in legitimately empty maps
//		zeoReportError("L3DTio_HFF::SaveTileFile error:\r\n - vert scale is invalid");
//		return false;
	}

	if(FH_HF300.TileSize>1) {
		if(FH_HF300.nx%FH_HF300.TileSize!=0 || FH_HF300.ny%FH_HF300.TileSize!=0) {
			zeoReportError("L3DTio_HFF::SaveTileFile error:\r\n - tile size is incompatible with map size in HFFv1.0");
			return false;
		}
	}
	
	// opening file for writing
	fh=fopen(lpFileName, "wb"); 
	if(fh==NULL) {
		TempStr.Format("L3DTio_HFF::SaveTileFile error:\r\n - cannot open file '%s' for writing", lpFileName);
		zeoReportError(TempStr);
		return false;
	}

	// writing header
	sprintf(TempBuf, "L3DT");
	fwrite(TempBuf, 4, 1, fh);
	fwrite((void*)&FileVersionNo, 2, 1, fh);
	sprintf(TempBuf, "HFF_v1.0");
	fwrite(TempBuf, 8, 1, fh);
	fwrite(&FH_HF300.DataOffset, 2, 1, fh);
	fwrite(&FH_HF300.nx, 4, 1, fh);
	fwrite(&FH_HF300.ny, 4, 1, fh);
	fwrite(&FH_HF300.DataSize, 1, 1, fh);
	fwrite(&FH_HF300.FloatFlag, 1, 1, fh);
	fwrite(&FH_HF300.VertScale, 4, 1, fh);
	fwrite(&FH_HF300.VertOffset, 4, 1, fh);
	fwrite(&FH_HF300.HorizScale, 4, 1, fh);
	fwrite(&FH_HF300.TileSize, 2, 1, fh);
	fwrite(&FH_HF300.WrapFlag, 1, 1, fh);
	TempBuf[0]=0;
	fwrite(TempBuf, 1, 23, fh);	// filling in reserved junk

	int TempTileSize = FH_HF300.TileSize;
	if(TempTileSize<1)
		TempTileSize = 1;
	
	bool ErrorFlag = false;

	int nty = ny / TempTileSize;
	int ntx = nx / TempTileSize;
	if(ny%TempTileSize)	nty++;
	if(nx%TempTileSize)	ntx++;

	for(ty=0; ty<nty && !ErrorFlag; ty++) {
	for(tx=0; tx<ntx && !ErrorFlag; tx++) {
		for(j=ty*TempTileSize; j<(ty+1)*TempTileSize && j<ny && !ErrorFlag; j++) {
		for(i=tx*TempTileSize; i<(tx+1)*TempTileSize && i<nx && !ErrorFlag; i++) {

			switch(MapTypeID) {
			default:
				ErrorFlag = true;
				break;
			case MAP_Heightfield:
				if(!zmap_tile_GetPixel(hTile, i, j, &HFpxl)) {
					zeoReportError("L3DTio_HFF::TileMapFile error:\r\n - GetPixel returned false for HF");
					ErrorFlag = true;
				}
				break;
			}

			if(!ErrorFlag) {
				if(FH_HF300.FloatFlag) {
					TempFloat=(HFpxl - FH_HF300.VertOffset)/FH_HF300.VertScale; // note gain & offset are taken into account when saving floats
					fwrite((void*)&TempFloat, 4, 1, fh);
				} else {
					switch(FH_HF300.DataSize) {
					case 1:
						TempInt=(int)((HFpxl-FH_HF300.VertOffset)/FH_HF300.VertScale);
						if(TempInt<0)		TempInt=0;
						if(TempInt>255)		TempInt=255;
						TempByte=(BYTE)TempInt;
						fwrite((void*)&TempByte, 1, 1, fh);
						break;
					case 2:
						TempInt=(int)((HFpxl-FH_HF300.VertOffset)/FH_HF300.VertScale);
						if(TempInt<0)		TempInt=0;
						if(TempInt>65535)	TempInt=65535;
						TempUSI=(USHORT)TempInt;
						fwrite((void*)&TempUSI, 2, 1, fh);
						break;
					default:
						zeoReportError("L3DTio_HFF::SaveTileFile error\r\n - pixel size is invalid");
						ErrorFlag = true;
						break;
					}
				}
			}
		}
		}
	}
	}


	fclose(fh);

	if(ErrorFlag) {
		zeoReportError("L3DTio_HFF::SaveTileFile error:\r\n - Save terminated following errors");
		return false;
	}
	
	return true;
}

zeoexport bool __stdcall zeoLoadTileFile(ZMAP hMap, LPVOID hTile, LPCSTR lpFileName, ZFORMAT hFormat) {

	CzMap zmap;
	zmap.Attach(hMap);

	CzFormat format;
	format.Attach(hFormat);
	
	struct _iobuf *fh;
	int i, j, tx, ty, i2, j2;
	char TempBuf[8];
	USHORT TempUSI;
	float TempFloat;
	CString TempStr;
	BYTE TempByte;
	char FileVersionCharBuf[9];
	char* CharBuf;


	USHORT FileVersionNo;

	// opening file for reading
	fh=fopen(lpFileName, "rb"); 
	if(fh==NULL) {
		TempStr.Format("L3DTio_HFF::LoadTileFile error\r\n - File '%s' not found fo reading", lpFileName);
		zeoReportError(TempStr);
		return false;
	}

	// check the file header
	fread(TempBuf, 4, 1, fh);
	TempBuf[4] = 0; 	// null-terminating string
	if(strcmp(TempBuf, "L3DT")!=0) {
		// case for old maps that didn't have "L3DT" written at start
		zeoReportError("L3DTio_HFF::LoadTileFile error\r\n - invalid file type");
		fclose(fh);
		return false;
	}
		
	// read the file version number
	fread(&FileVersionNo, 2, 1, fh);
	switch(FileVersionNo) {
	default:
		zeoReportError("L3DTio_HFF::LoadTileFile error\r\n - unknown file version");
		fclose(fh);
		return false;
	case DEF_FILE_HF300:
		//this is OK
		break;
	}

	// reading ASCII external file version number
	fread(FileVersionCharBuf, 8, 1, fh);
	FileVersionCharBuf[8]=0;	// null-terminating string
	if(strcmp(FileVersionCharBuf, "HFF_v1.0")!=0) {
		TempStr.Format("L3DTio_HFF::LoadTileFile error\r\n - unknown file version '%s' in HF300", FileVersionCharBuf);
		zeoReportError(TempStr);
		fclose(fh);
		return false;
	}

	// declare a file header struct
	FileHeader_HF300 FH_HF300;

	// read the data offset
	fread(&FH_HF300.DataOffset, 2, 1, fh);

	// check data offset
	if(FH_HF300.DataOffset<64) {
		zeoReportError("L3DTio_HFF::LoadTileFile error:\r\n - Invalid file header length");
		fclose(fh);
		return false;
	}

	// reading remainder of header into char buffer
	CharBuf=(char*)calloc(1, FH_HF300.DataOffset-16);
	fread(CharBuf, FH_HF300.DataOffset-16, 1, fh);

	// copy from buffer into header struct
	memcpy(&FH_HF300.nx, CharBuf, 4);
	memcpy(&FH_HF300.ny, CharBuf+4, 4);
	memcpy(&FH_HF300.DataSize, CharBuf+8, 1);
	memcpy(&FH_HF300.FloatFlag, CharBuf+9, 1);
	memcpy(&FH_HF300.VertScale, CharBuf+10, 4);
	memcpy(&FH_HF300.VertOffset, CharBuf+14, 4);
	memcpy(&FH_HF300.HorizScale, CharBuf+18, 4);
	memcpy(&FH_HF300.TileSize, CharBuf+22, 2);
	memcpy(&FH_HF300.WrapFlag, CharBuf+24, 1);

	// free buffer
	free(CharBuf);

	// check tiling (is it compatible with map size?)
	if(FH_HF300.TileSize>1) {
		if(FH_HF300.nx%FH_HF300.TileSize!=0 || FH_HF300.ny%FH_HF300.TileSize!=0) {
			zeoReportError("L3DTio_HFF::LoadTileFile error\r\n - tile size is incompatible with map size in HF300");
			fclose(fh);
			return false;
		}
	}

	// check vertical scaling
	if(FH_HF300.VertScale<=0.0) {
		FH_HF300.VertScale = 1.0f;
		//zeoReportError("L3DTio_HFF::LoadTileFile error\r\n - vertical scale is equal to or smaller than zero");
		//fclose(fh);
		//return false;
	}

	// check horizontal scaling
	if(FH_HF300.HorizScale<=0.0) {
		zeoReportError("L3DTio_HFF::LoadTileFile error\r\n - horizontal scale is equal to or smaller than zero");
		fclose(fh);
		return false;
	}

	// checking data types
	if(FH_HF300.FloatFlag) {
		switch(FH_HF300.DataSize) {
		case 4:	// that's OK
			break;
		default:
			zeoReportError("L3DTio_HFF::LoadTileFile error\r\n - unacceptable data size (float)");
			fclose(fh);
			return false;
		}
	} else {
		switch(FH_HF300.DataSize) {
		case 1: // BYTE is OK
			break;
		case 2: // USHORT is OK
			break;
		default:
			zeoReportError("L3DTio_HFF::LoadTileFile error\r\n - unacceptable data size (int)");
			fclose(fh);
			return false;
		}
	}

	long nx = FH_HF300.nx;
	long ny = FH_HF300.ny;

	if(nx<=0||ny<=0 || nx!=ny) {
		zeoReportError("L3DTio_HFF::LoadTileFile error\r\n - invalid map size");
		fclose(fh);
		return false;
	}

	if(nx!=zmap.GetTileSize()) {
		zeoReportError("L3DTio_HFF::LoadTileFile error\r\n - file size does not match tile size");
		fclose(fh);
		return false;
	}


	// reading data (handling both tiled and non-tiled maps)
	int TileSize = FH_HF300.TileSize;
	if(TileSize<=0)	TileSize = 1;

	// stepping through tiles
	bool ErrorFlag = false;
	for(ty=0; ty<ny/TileSize && !ErrorFlag; ty++) {
	for(tx=0; tx<nx/TileSize && !ErrorFlag; tx++) {

		// stepping through pixels in each tile
		for(j=0; j<TileSize && !ErrorFlag; j++) {
		for(i=0; i<TileSize && !ErrorFlag; i++) {

			// converting from local tile coords to global map coords
			i2 = i+tx*TileSize;
			j2 = j+ty*TileSize;

			if(FH_HF300.FloatFlag) {
				fread((void*)&TempFloat, 4, 1, fh);
				TempFloat = TempFloat*FH_HF300.VertScale+FH_HF300.VertOffset;
			} else {
				switch(FH_HF300.DataSize) {
				case 1:
					fread((void*)&TempByte, 1, 1, fh);
					TempFloat = (float)TempByte*FH_HF300.VertScale+FH_HF300.VertOffset;
					break;	
				case 2:
					fread((void*)&TempUSI, 2, 1, fh);
					TempFloat = (float)TempUSI*FH_HF300.VertScale+FH_HF300.VertOffset;
					break;
				default:
					ErrorFlag = true;
					zeoReportError("L3DTio_HFF::LoadMapFile error\n - pixel size is invalid in HF300");
					break;
				}
			}

			if(!zmap_tile_SetPixel(hTile, i2, j2, &TempFloat)) {
				ErrorFlag = true;
				zeoReportError("L3DTio_HFF::LoadTileFile error\n - call to SetPixel failed");
			}
		}
		}
	}
	}

	fclose(fh);

	if(ErrorFlag) {
		return false;
	}

	// updating file config	
	int datasize = FH_HF300.DataSize;
	int tilesize = FH_HF300.TileSize;
	bool floatflag = FH_HF300.FloatFlag;
	format.SetOptionValue("DataSize", VarID_int, &datasize);
	format.SetOptionValue("TileSize", VarID_int, &tilesize);
	format.SetOptionValue("FloatFlag", VarID_bool, &floatflag);

	return true;
}